#include "myutils.h"
#include "seqdb.h"
#include "chsim.h"
#include "seq.h"
#include "path.h"
#include "dp.h"
#include "alpha.h"
#include <set>

#define TRACE		0

float GetAbFromLabel(const string &Label);
unsigned randu32();
void InitRand(unsigned Seed);
double GetFractIdGivenPath(const byte *A, const byte *B, const char *Path);

static FILE *g_fOut;
static FILE *g_fAccOut;
static unsigned g_WordLength;
static WordInfo **g_WordIndex;
static unsigned *g_WordCounts;
static float *g_WordAbs;
static unsigned g_WordCount;
static SeqDB *g_ptrSpeciesDB;
static vector<bool> g_UsedAsParentVec;
static unsigned g_SpecCount;
static unsigned g_SegSameRejectCount;
static unsigned g_AcceptCount;

void Usage()
	{
	printf("\n");
	printf("For more information, please visit:\n");
	printf("http://drive5.com/chsim\n");
	return;
	}

/***
Species label:
	sp<n>/ab=<x>/name=...

Chimera label:
	ch<n>/ab=<x>/<segs>

<segs>:
	<seg>/<seg>/.../<seg>

<seg>:
	sp<n>:<lo>-<hi>
***/

float GetAbFromLabel(const string &Label)
	{
	vector<string> Fields;
	Split(Label, Fields, '/');
	const unsigned N = SIZE(Fields);
	for (unsigned i = 0; i < N; ++i)
		{
		const string &Field = Fields[i];
		if (Field.substr(0, 3) == "ab=")
			{
			string a = Field.substr(3, string::npos);
			return (float) atof(a.c_str());
			}
		}
	return 0.0;
	}

static void ParseSeg(const string &s, ChSeg &Seg)
	{
	unsigned lo;
	unsigned hi;
	int n = sscanf(s.c_str(), "sp%u:%u-%u", &Seg.SeqIndex, &lo, &hi);
	if (n != 3)
		Die("Invalid seg: '%s'", s.c_str());

	Seg.SpecPos = lo;
	Seg.Length = hi - lo + 1;
	}

void ParseSpLabel(const string &Label, unsigned &Id, string &Name, float &Ab)
	{
	Ab = -1.0f;

	if (Label.substr(0, 2) != "sp")
		Die("Invalid species label, must start with sp %s", Label.c_str());

	const char *sId = Label.c_str() + 2;
	Id = 0;
	for (const char *p = sId; ; ++p)
		{
		char c = *p;
		if (c == '/' || c == 0)
			break;
		if (!isdigit(c))
			Die("Invalid label, id not integer: %s", Label.c_str());
		Id = Id*10 + (c - '0');
		}

	vector<string> Fields;
	Split(Label, Fields, '/');

	const unsigned N = SIZE(Fields);
	unsigned ChPos = 0;
	ChSeg Seg;
	Seg.SeqIndex = UINT_MAX;
	for (unsigned i = 1; i < N; ++i)
		{
		const string &Field = Fields[i];
		if (Field.substr(0, 3) == string("ab="))
			Ab = (float) atof(Field.c_str() + 3);
		else if (Field.substr(0, 5) == string("name="))
			Name = Field.substr(5, string::npos);
		}
	}

void ParseChLabel(const string &Label, bool &IsSpecies, unsigned &Id, string &Name,
  float &Ab, vector<ChSeg> &Segs)
	{
	Segs.clear();
	Ab = -1.0f;

	if (Label.substr(0, 2) == "sp")
		IsSpecies = true;
	else if (Label.substr(0, 2) == "ch")
		IsSpecies = false;
	else
		Die("Invalid label, must start with sp or ch: %s", Label.c_str());

	const char *sId = Label.c_str() + 2;
	Id = 0;
	for (const char *p = sId; ; ++p)
		{
		char c = *p;
		if (c == '/' || c == 0)
			break;
		if (!isdigit(c))
			Die("Invalid label, id not integer: %s", Label.c_str());
		Id = Id*10 + (c - '0');
		}

	vector<string> Fields;
	Split(Label, Fields, '/');

	const unsigned N = SIZE(Fields);
	unsigned ChPos = 0;
	ChSeg Seg;
	Seg.SeqIndex = UINT_MAX;
	for (unsigned i = 1; i < N; ++i)
		{
		const string &Field = Fields[i];
		if (Field.substr(0, 3) == string("ab="))
			Ab = (float) atof(Field.c_str() + 3);
		else if (Field.substr(0, 2) == string("sp"))
			{
			if (IsSpecies)
				Die("Seg field invalid in species: %s", Label.c_str());
			Seg.ChPos = ChPos;
			ParseSeg(Field.c_str(), Seg);
			Segs.push_back(Seg);
			ChPos += Seg.Length;
			}
		else if (Field.substr(0, 5) == string("name="))
			{
			if (!IsSpecies)
				Die("Name field invalid in chimera label: %s", Label.c_str());
			Name = Field.substr(5, string::npos);
			}
		else
			; // Die("Invalid field '%s' in label: %s", Field.c_str(), Label.c_str());
		}

	if (Ab <= 0.0f)
		Die("Missing or invalid abundance field: %s", Label.c_str());
	}

static unsigned GetRandDist(const float *w, unsigned N)
	{
	float Sumw = 0.0f;
	for (unsigned i = 0; i < N; ++i)
		Sumw += w[i];

	unsigned r = randu32();
	float s = Sumw*float(r)/float(UINT_MAX);

	float Sumw2 = 0.0f;
	for (unsigned i = 0; i < N; ++i)
		{
		Sumw2 += w[i];
		if (Sumw2 >= s && w[i] > 0.0f)
			return i;
		}
	return N-1;
	}

#if	0
void TestRandDist()
	{
	const unsigned M = 1000000;

	float *w = myalloc(float, 1000);
	unsigned *c = myalloc(unsigned, 1000);
	zero(c, 1000);

	w[0] = 0.3f;
	w[1] = 10.0f;
	w[2] = 3.0f;
	w[3] = 5.0f;
	unsigned N = 4;

	float Sumw = 0.0f;
	for (unsigned i = 0; i < N; ++i)
		Sumw += w[i];

	for (unsigned i = 0; i < M; ++i)
		{
		unsigned k = GetRandDist(w, N);
		++(c[k]);
		}

	Log("M=%u\n", M);
	for (unsigned i = 0; i < N; ++i)
		{
		float p = w[i]/Sumw;
		float f = float(c[i])/M;
		Log("%2u: w %.1f, p %.3g, c %u, f %.3g\n",
		  i, w[i], w[i]/Sumw, c[i], f);
		}

	myfree(w);
	myfree(c);
	}
#endif // 0

static void InitWordIndex(unsigned WordLength)
	{
	if (g_WordCounts != 0)
		{
		for (unsigned Word = 0; Word < g_WordCount; ++Word)
			{
			WordInfo *Next;
			for (WordInfo *WI = g_WordIndex[Word]; WI; WI = Next)
				{
				Next = WI->Next;
				delete WI;
				}
			}
		myfree(g_WordCounts);
		myfree(g_WordAbs);
		myfree(g_WordIndex);
		}

	g_WordLength = WordLength;
	g_WordCount = 1;
	for (unsigned i = 0; i < WordLength; ++i)
		g_WordCount *= 4;

	g_WordCounts = myalloc(unsigned, g_WordCount);
	g_WordAbs = myalloc(float, g_WordCount);
	g_WordIndex = myalloc(WordInfo *, g_WordCount*sizeof(WordInfo *));

	zero(g_WordCounts, g_WordCount);
	zero(g_WordAbs, g_WordCount);
	zero(g_WordIndex, g_WordCount);
	}

static void AddWord(unsigned Word, unsigned SeqIndex, unsigned Pos, float SeqAb)
	{
	asserta(Word < g_WordCount);

	++(g_WordCounts[Word]);
	g_WordAbs[Word] += SeqAb;

	WordInfo *WI = new WordInfo;
	asserta(WI != 0);

	WI->SeqIndex = SeqIndex;
	WI->Pos = Pos;
	WI->Next = g_WordIndex[Word];

	g_WordIndex[Word] = WI;
	}

static word_t *g_Words = 0;
static unsigned g_WordsCapacity = 0;

static unsigned GetWord(const byte *Seq)
	{
	word_t Word = 0;
	const byte *Front = Seq;
	for (unsigned i = 0; i < g_WordLength; ++i)
		{
		unsigned Letter = g_CharToLetterNucleo[*Front++];
		Word = (Word*4) + Letter;
		}
	return Word;
	}

static void AddSeq(unsigned SeqIndex, const byte *Seq, unsigned L, float Ab)
	{
	if (L <= g_WordLength || L <= opt_flank)
		return;

	for (unsigned Pos = opt_flank; Pos + g_WordLength < L - opt_flank; ++Pos)
		{
		unsigned Word = GetWord(Seq+Pos);
		AddWord(Word, SeqIndex, Pos, Ab);
		}
	}

static bool GetLocusPair(unsigned WordX, const vector<float> &SeqAbs,
  WordInfo &W1, WordInfo &W2)
	{
	const unsigned N = g_WordCounts[WordX];
	asserta(N > 1);

	float *w = myalloc(float, N);
	unsigned k = 0;
	for (const WordInfo *WI = g_WordIndex[WordX]; WI; WI = WI->Next)
		{
		unsigned SeqIndex = WI->SeqIndex;
		w[k++] = SeqAbs[SeqIndex];
		}
	asserta(k == N);

	unsigned i1 = GetRandDist(w, N);
	asserta(w[i1] > 0.0f);
	w[i1] = 0.0f;
	unsigned i2 = GetRandDist(w, N);
	asserta(i1 != i2);
	myfree(w);

	const WordInfo *WI1 = g_WordIndex[WordX];
	const WordInfo *WI2 = g_WordIndex[WordX];
	for (unsigned i = 0; i < i1; ++i)
		WI1 = WI1->Next;
	for (unsigned i = 0; i < i2; ++i)
		WI2 = WI2->Next;
#if	TRACE
	Log("  Seq pos: %u,%u and %u,%u\n",
		WI1->SeqIndex, WI1->Pos,
		WI2->SeqIndex, WI2->Pos);
	if (WI1->SeqIndex == WI2->SeqIndex)
		Log(" Non-homologous cross!\n");
#endif

// Most abundant usually picked first, randomize order
	if (randu32()%2 == 0)
		{
		W1 = *WI1;
		W2 = *WI2;
		}
	else
		{
		W1 = *WI2;
		W2 = *WI1;
		}
	return true;
	}

static void SegsToLabel(const vector<ChSeg> &Segs, unsigned lo,
  unsigned hi, string &Label)
	{
	Label.clear();
	const unsigned N = SIZE(Segs);
	for (unsigned i = 0; i < N; ++i)
		{
		const ChSeg &Seg = Segs[i];
		unsigned SeqIndex = Seg.SeqIndex;

		unsigned ChLo = Seg.ChPos;
		unsigned ChHi = Seg.ChPos + Seg.Length - 1;
		if (ChLo > hi)
			return;

		if (ChHi < lo)
			continue;

		unsigned SpecLo = Seg.SpecPos;
		unsigned SpecHi = Seg.SpecPos + Seg.Length - 1;

		asserta(SeqIndex < g_ptrSpeciesDB->GetSeqCount());
		asserta(SpecLo <= SpecHi);
		asserta(SpecHi < g_ptrSpeciesDB->GetSeqLength(SeqIndex));

		if (ChLo < lo)
			SpecLo += lo - ChLo;
		if (ChHi > hi)
			SpecHi -= ChHi - hi;

		asserta(SpecLo <= SpecHi);
		asserta(SpecHi < g_ptrSpeciesDB->GetSeqLength(SeqIndex));

		char s[64];
		sprintf(s, "sp%u:%u-%u", SeqIndex, SpecLo, SpecHi);

		if (!Label.empty())
			Label.push_back('/');
		Label += string(s);
		}
	}

static void GetSegLabel(const string &Label, unsigned lo, unsigned hi,
  string &SegLabel)
	{
	SegLabel.clear();

	bool IsSpecies;
	unsigned Id;
	string Name;
	float Ab;
	vector<ChSeg> Segs;
	ParseChLabel(Label, IsSpecies, Id, Name, Ab, Segs);
	if (IsSpecies)
		{
		char s[64];
		sprintf(s, "sp%u:%u-%u", Id, lo, hi);
		asserta(Id < g_ptrSpeciesDB->GetSeqCount());
		asserta(lo <= hi);
		asserta(hi < g_ptrSpeciesDB->GetSeqLength(Id));
		SegLabel = string(s);
		return;
		}

	SegsToLabel(Segs, lo, hi, SegLabel);
	}

static float GetChAbNoise()
	{
	if (opt_chsim_abnoisepct == 0)
		return 1.0;

	unsigned i = randu32()%100;
	unsigned Pct = randu32()%opt_chsim_abnoisepct;
	if (i%2 == 0)
		return 1.0f + Pct/100.0f;
	else
		return 1.0f - Pct/100.0f;
	}

static void Validate(const string &Label, const byte *ChSeq, unsigned ChL, 
  unsigned &SegCount, const SeqDB &SpeciesDB)
	{
	SegCount = 0;
	bool IsSpecies;
	unsigned Id;
	string Name;
	float Ab;
	vector<ChSeg> Segs;
	ParseChLabel(Label, IsSpecies, Id, Name, Ab, Segs);
	if (IsSpecies)
		return;

	const unsigned N = SIZE(Segs);
	SegCount = N;
	for (unsigned i = 0; i < N; ++i)
		{
		const ChSeg &Seg = Segs[i];
		unsigned SeqIndex = Seg.SeqIndex;
		unsigned SpecLo = Seg.SpecPos;
		unsigned Length = Seg.Length;
		unsigned SpecHi = SpecLo + Length - 1;

		const byte *SpecSeq = SpeciesDB.GetSeq(SeqIndex);
		unsigned SpecL = SpeciesDB.GetSeqLength(SeqIndex);
		
		asserta(SpecLo < SpecL);
		asserta(SpecHi < SpecL);
		asserta(SpecLo <= SpecHi);

		unsigned ChPos = Seg.ChPos;
		for (unsigned SpecPos = SpecLo; SpecPos <= SpecHi; ++SpecPos, ++ChPos)
			{
			asserta(ChPos < ChL);
			asserta(SpecPos < SpecL);
			asserta(ChSeq[ChPos] == SpecSeq[SpecPos]);
			}
		}
	}

static void GetDiv(const string &Label, const byte *ChSeq, unsigned ChL,
  float &TopPctId, unsigned &TopSeqIndex)
	{
	TopSeqIndex = UINT_MAX;
	TopPctId = -1.0f;

	const SeqDB &SpeciesDB = *g_ptrSpeciesDB;

	bool IsSpecies;
	unsigned Id;
	string Name;
	float Ab;
	vector<ChSeg> Segs;
	ParseChLabel(Label, IsSpecies, Id, Name, Ab, Segs);
	asserta(!IsSpecies);

	SeqData Query;
	Query.Label = Label.c_str();
	Query.Seq = ChSeq;
	Query.L = ChL;

	const unsigned N = SIZE(Segs);
	unsigned TopParentSeqIndex = UINT_MAX;
	Log(" Query %s\n", Query.Label);
	for (unsigned i = 0; i < N; ++i)
		{
		const ChSeg &Seg = Segs[i];
		unsigned SeqIndex = Seg.SeqIndex;

		SeqData Target;
		SpeciesDB.GetSeqData(SeqIndex, Target);

		PathData PD;
		GlobalAlign(Query, Target, PD);

		double Id = GetFractIdGivenPath(Query.Seq, Target.Seq, PD.Start);
		Log("  Div %6.1f%%  %s\n", Id*100.0, Target.Label);
		if (Id*100.0 > TopPctId)
			{
			TopPctId = float(100.0*Id);
			TopSeqIndex = SeqIndex;
			}
		}
	}

static bool SegFound(const byte *Seg, unsigned SegL, const byte *Seq, unsigned L)
	{
	for (unsigned i = 0; i < L; ++i)
		{
		for (unsigned j = 0; j < SegL; ++j)
			{
			if (i + j >= L)
				return false;
			if (Seg[j] != Seq[i+j])
				goto Nexti;
			}
		return true;
	Nexti:;
		}
	return false;
	}

static bool MakeChimera(unsigned ChIx, const SeqDB &InDB, const vector<float> &InAbs, 
  const WordInfo &WI1, const WordInfo &WI2, float ChAbFactor, SeqData &Ch, float &ChAb,
  float &TopPctId)
	{
	unsigned SeqIndex1 = WI1.SeqIndex;
	unsigned SeqIndex2 = WI2.SeqIndex;

	unsigned Pos1 = WI1.Pos;
	unsigned Pos2 = WI2.Pos;

	const byte *Seq1 = InDB.GetSeq(SeqIndex1);
	const byte *Seq2 = InDB.GetSeq(SeqIndex2);

	const char *Label1 = InDB.GetLabel(SeqIndex1);
	const char *Label2 = InDB.GetLabel(SeqIndex2);

	float Ab1 = InAbs[SeqIndex1];
	float Ab2 = InAbs[SeqIndex2];

	const unsigned L1 = InDB.GetSeqLength(SeqIndex1);
	const unsigned L2 = InDB.GetSeqLength(SeqIndex2);

	for (unsigned i = 0; i < g_WordLength; ++i)
		{
		char c1 = Seq1[Pos1+i];
		char c2 = Seq2[Pos2+i];
		asserta(c1 == c2);
		}

	if (opt_chsim_chab == 0.0f)
		ChAb = Ab1*Ab2*ChAbFactor*GetChAbNoise();
	else
		ChAb = float(opt_chsim_chab);
	
	unsigned w2 = g_WordLength/2;

	unsigned hi1 = Pos1 + w2;
	unsigned lo2 = Pos2 + w2 + 1;

	unsigned SegLen1 = hi1 + 1;
	unsigned SegLen2 = L2 - lo2;
	unsigned ChL = SegLen1 + SegLen2;

	byte *ChSeq = myalloc(byte, ChL);
	memcpy(ChSeq, Seq1, SegLen1);
	memcpy(ChSeq + SegLen1, Seq2 + lo2, SegLen2);

	string SegLabel1;
	string SegLabel2;
	GetSegLabel(Label1, 0, hi1, SegLabel1);
	GetSegLabel(Label2, lo2, L2-1, SegLabel2);

	char chn[64];
	sprintf(chn, "ch%u/ab=%.3g/", ChIx, ChAb); 
	string SegLabel = string(chn) + SegLabel1 + string("/") + SegLabel2;

	bool IsSpecies;
	unsigned Id;
	string Name;
	float Ab;
	vector<ChSeg> Segs2;
	ParseChLabel(SegLabel, IsSpecies, Id, Name, Ab, Segs2);

	const unsigned SegCount2 = SIZE(Segs2);
	vector<ChSeg> Segs;
	for (unsigned i = 0; i < SegCount2; ++i)
		{
		if (i + 1 < SegCount2 && Segs2[i].SeqIndex == Segs2[i+1].SeqIndex
			&& Segs2[i].SpecPos + Segs2[i].Length == Segs2[i+1].SpecPos)
			{
			ChSeg MergedSeg;
			MergedSeg.SeqIndex = Segs2[i].SeqIndex;
			MergedSeg.ChPos = Segs2[i].ChPos;
			MergedSeg.SpecPos = Segs2[i].SpecPos;
			MergedSeg.Length = Segs2[i].Length + Segs2[i+1].Length;
			Segs.push_back(MergedSeg);
			++i;
			}
		else
			Segs.push_back(Segs2[i]);
		}

	const unsigned SegCount = SIZE(Segs);

	for (unsigned i = 0; i < SegCount; ++i)
		{
		const ChSeg &Seg = Segs[i];
		unsigned ParentSeqIndex = Seg.SeqIndex;
		const byte *ParentSeq = g_ptrSpeciesDB->GetSeq(ParentSeqIndex);
		unsigned ParentL = g_ptrSpeciesDB->GetSeqLength(ParentSeqIndex);
		asserta(Seg.SpecPos + Seg.Length <= ParentL);
		const byte *SegSeq = ParentSeq + Seg.SpecPos;
		unsigned SegSeqL = Seg.Length;

		for (unsigned j = 0; j < SegCount; ++j)
			{
			const ChSeg &Segj = Segs[j];
			unsigned ParentSeqIndexj = Segj.SeqIndex;
			if (ParentSeqIndexj == ParentSeqIndex)
				continue;
			const byte *ParentSeqj = g_ptrSpeciesDB->GetSeq(ParentSeqIndexj);
			unsigned ParentLj = g_ptrSpeciesDB->GetSeqLength(ParentSeqIndexj);
			bool Found = SegFound(SegSeq, SegSeqL, ParentSeqj, ParentLj);
			//Log("\n");
			//Log("SegFound=%c\n", tof(Found));
			//Log("sp%u %*.*s\n", ParentSeqIndex, ParentL, ParentL, ParentSeq);
			//Log("sp%u %*.*s\n", ParentSeqIndexj, ParentLj, ParentLj, ParentSeqj);
			//Log("Seg %*.*s\n", SegSeqL, SegSeqL, SegSeq);
			if (Found)
				{
				++g_SegSameRejectCount;
				return false;
				}
			}
		}

	sprintf(chn, "ch%u/ab=%.3g/", ChIx, ChAb);
	string Label = string(chn);

	string SegLabel3;
	SegsToLabel(Segs, 0, UINT_MAX, SegLabel3);
	Label += SegLabel3;

	unsigned LabelLen = (unsigned) strlen(Label.c_str());
	unsigned ChLabelLen = LabelLen + 1024;
	char *ChLabel = myalloc(char, ChLabelLen);
	strcpy(ChLabel, Label.c_str());

	Ch.Seq = ChSeq;
	Ch.L = ChL;
	Ch.Label = ChLabel;

	unsigned SegCountv;
	Validate(Ch.Label, Ch.Seq, Ch.L, SegCountv, *g_ptrSpeciesDB);

	unsigned TopSeqIndex;
	GetDiv(Ch.Label, Ch.Seq, Ch.L, TopPctId, TopSeqIndex);

	char s[128];
	sprintf(s, "/N=%u/top=sp%u:%.1f%%", SegCount, TopSeqIndex, TopPctId);

	strcat(ChLabel, s);

	if (SegCount >= opt_chsim_minm && SegCount >= opt_chsim_maxm &&
	  TopPctId >= opt_chsim_divlo && TopPctId < opt_chsim_divhi)
		{
		++g_AcceptCount;
		if (g_fAccOut != 0)
			{
			fprintf(g_fAccOut, ">%s\n", Ch.Label);
			fprintf(g_fAccOut, "%*.*s\n", Ch.L, Ch.L, Ch.Seq);
			}
		if (g_AcceptCount >= opt_chsim_acc)
			{
			Progress("\n");
			Progress("%u accepts\n", g_AcceptCount);
			CloseStdioFile(g_fOut);
			CloseStdioFile(g_fAccOut);
			ProgressExit();
			exit(0);
			}
		return false;
		}

	return true;
	}

static void MakePStr(const char *Label, char *PStr)
	{
	bool IsSpecies;
	string Name;
	float Ab;
	unsigned Id;
	vector<ChSeg> Segs;
	ParseChLabel(Label, IsSpecies, Id, Name, Ab, Segs);

	if (IsSpecies)
		sprintf(PStr, "sp%u", Id);
	else
		sprintf(PStr, "Ch%u", Id);
	}

void ChSimIter(unsigned Iter, float ChAbFactor, unsigned &ChIx, const SeqDB &InDB,
  const vector<float> &InAbs, SeqDB &OutDB, vector<float> &OutAbs)
	{
	InitWordIndex(opt_k);

	const unsigned InSeqCount = InDB.GetSeqCount();
	asserta(SIZE(InAbs) == InSeqCount);

	for (unsigned SeqIndex = 0; SeqIndex < InSeqCount; ++SeqIndex)
		{
		const char *Label = InDB.GetLabel(SeqIndex);
		const byte *Seq = InDB.GetSeq(SeqIndex);
		unsigned L = InDB.GetSeqLength(SeqIndex);
		float Ab = InAbs[SeqIndex];
		AddSeq(SeqIndex, Seq, L, Ab);

		OutDB.AddSeq(Label, Seq, L);
		OutAbs.push_back(Ab);
		}

	float SumAbs = 0;
	for (unsigned Word = 0; Word < g_WordCount; ++Word)
		{
		if (g_WordCounts[Word] > 1)
			{
			SumAbs += g_WordAbs[Word];
			if (SumAbs > UINT_MAX/2)
				Die("SubAbs too big");
			}
		}

#if	TRACE
	Log("SumAbs=%u\n", SumAbs);
#endif

	asserta(SumAbs > 0);

	for (unsigned Word = 0; Word < g_WordCount; ++Word)
		if (g_WordCounts[Word] < 2)
			g_WordAbs[Word] = 0.0f;

	Log("\n");
	Log("ChAbFactor %.3g\n", ChAbFactor);
	Log("Iter   Chix       P1       P2   Pos1   Pos2  ChLen       Ab1         Ab2        ChAb        Word  Label\n");
	Log("----  -----  -------  -------  -----  -----  -----  ----------  ----------  ----------  ----------  -----\n");

	for (unsigned Chix = 0; Chix < opt_chsim_nperiter; ++Chix)
		{
#if	TRACE
		Log("\n");
		Log("Chix=%u\n", Chix);
#endif

		SeqData Ch;
		float ChAb;
		float TopPctId;
		unsigned WordX;
		WordInfo WI1, WI2;
		for (;;)// unsigned Try2 = 0; Try2 < 16; ++Try2)
			{
		// Choose cross-over word
			WordX = GetRandDist(g_WordAbs, g_WordCount);
			asserta(WordX != UINT_MAX);
#if	TRACE
			Log("Word=%s abs=%.1f\n", WordToStr(WordX, g_WordLength, true),
			  g_WordAbs[WordX]);
#endif
			GetLocusPair(WordX, InAbs, WI1, WI2);

			bool Ok = MakeChimera(ChIx, InDB, InAbs, WI1, WI2, ChAbFactor, Ch, ChAb, TopPctId);
			if (!Ok)
				continue;
			if (TopPctId < 100.0)
				{
				if (WI1.SeqIndex < g_SpecCount)
					g_UsedAsParentVec[WI1.SeqIndex] = true;

				if (WI2.SeqIndex < g_SpecCount)
					g_UsedAsParentVec[WI2.SeqIndex] = true;

				break;
				}
			}
		if (TopPctId == 100.0)
			Warning("Failed to make 100.0");

		const char *Label1 = InDB.GetLabel(WI1.SeqIndex);
		const char *Label2 = InDB.GetLabel(WI2.SeqIndex);

		char P1s[64];
		char P2s[64];
		MakePStr(Label1, P1s);
		MakePStr(Label2, P2s);

		Log("%4u  %5u  %7.7s  %7.7s  %5u  %5u  %5u  %10.3g  %10.3g  %10.3g  %s  %s\n",
		  Iter,
		  ChIx,
		  P1s,
		  P2s,
		  WI1.Pos, WI2.Pos,
		  Ch.L,
		  InAbs[WI1.SeqIndex], InAbs[WI2.SeqIndex], ChAb,
		  WordToStr(WordX, g_WordLength, true),
		  Ch.Label);

		OutDB.AddSeq(Ch.Label, Ch.Seq, Ch.L);
		OutAbs.push_back(ChAb);

		asserta(OutDB.GetSeqCount() == SIZE(OutAbs));
		++ChIx;

		myfree((void *) Ch.Seq);
		}
	}

void ChSim()
	{
	if (opt_input == "")
		Die("Missing --input");
	if (opt_output == "" && opt_accout == "")
		Die("Must specify --output and/or --accout");

	FILE *fOut = 0;
	if (opt_output != "")
		fOut= CreateStdioFile(opt_output);
	g_fOut = fOut;
	
	if (opt_accout != "")
		g_fAccOut = CreateStdioFile(opt_accout);

	if (!optset_k)
		opt_k = 10;

	InitRand(opt_randseed);

	SeqDB Input;
	Input.FromFasta(opt_input);
	Input.SetIsNucleo();
	g_ptrSpeciesDB = &Input;
	const unsigned SpecSeqCount = Input.GetSeqCount();

	vector<float> SpecAbs;
	SpecAbs.resize(SpecSeqCount, 2.0f);
	if (opt_chsim_expab > 0)
		{
		unsigned K = SpecSeqCount;
		for (float Ab = float(opt_chsim_expab); Ab > 2.0f; Ab /= 2.0f)
			{
			if (K == 0)
				break;
			unsigned r = randu32()%K;
			for (unsigned SpecIndex = r; SpecIndex < SpecSeqCount; ++SpecIndex)
				if (SpecAbs[SpecIndex] == 2.0f)
					{
					SpecAbs[SpecIndex] = Ab;
					break;
					}
			--K;
			}
		}

	for (unsigned SpecIndex = 0; SpecIndex < SpecSeqCount; ++SpecIndex)
		{
		float Ab;
		const char *Label = Input.GetLabel(SpecIndex);
		if (opt_chsim_expab == 0)
			Ab = GetAbFromLabel(string(Label));
		else
			{
			string Name;
			unsigned Id;
			vector<ChSeg> Segs;
			bool IsSpecies;
			ParseChLabel(Label, IsSpecies, Id, Name, Ab, Segs);
			if (!IsSpecies)
				Die("Chimera in input file: %s\n", Label);
			if (Id != SpecIndex)
				Die("Species not numbered sequentially from zero: %s\n", Label);
			}
		SpecAbs[SpecIndex] = Ab;
		}

	g_SpecCount = SpecSeqCount;
	g_UsedAsParentVec.clear();
	g_UsedAsParentVec.resize(SpecSeqCount, false);

	SeqDB Output;
	vector<float> OutAbs;

	SeqDB *ptrInDB = &Input;
	SeqDB *ptrOutDB = &Output;

	vector<float> *ptrInAbs = &SpecAbs;
	vector<float> *ptrOutAbs = &OutAbs;
	unsigned Iter = 0;
	unsigned ChIx = 0;
	float ChAbFactor = (float) (opt_chsim_abfactor*pow(2.0, double(opt_chsim_iters-1)));
	for (;;)
		{
		ProgressStep(Iter, opt_chsim_iters, "Iter %u, %u accepts", Iter, g_AcceptCount);

		ptrOutDB->Clear();
		ptrOutAbs->clear();
		ptrOutDB->m_IsNucleo = true;
		ptrOutDB->m_IsNucleoSet = true;
		ChSimIter(Iter, ChAbFactor, ChIx, *ptrInDB, *ptrInAbs, *ptrOutDB, *ptrOutAbs);
		Log("\n");
		Log("Same seg rejects %u\n", g_SegSameRejectCount);
		Log("\n");
		if (++Iter == opt_chsim_iters)
			break;

		unsigned N = ptrOutDB->GetSeqCount();
		for (unsigned i = 0; i < N; ++i)
			{
			string Label = ptrOutDB->GetLabel(i);
			const byte *Seq = ptrOutDB->GetSeq(i);
			unsigned L = ptrOutDB->GetSeqLength(i);
			unsigned SegCount;
			Validate(Label, Seq, L, SegCount, Input);
			}
		
		swap(ptrInDB, ptrOutDB);
		swap(ptrInAbs, ptrOutAbs);

		ChAbFactor /= 2.0f;
		}

	ptrOutDB->ToFasta(opt_output);

	const unsigned OutSeqCount = ptrOutDB->GetSeqCount();
	for (unsigned SeqIndex = 0; SeqIndex < OutSeqCount; ++SeqIndex)
		{
		void SeqToFasta(FILE *f, const char *Label, const byte *Seq, unsigned L);
		const char *Label = ptrOutDB->GetLabel(SeqIndex);
		string sLabel = string(Label);
		if (strncmp(Label, "sp", 2) == 0)
			{
			string Name;
			float Ab;
			unsigned Id;
			ParseSpLabel(sLabel, Id, Name, Ab);
			asserta(Id == SeqIndex);

			Ab = (*ptrOutAbs)[SeqIndex];
			if (!g_UsedAsParentVec[Id])
				Ab = 1.0f;

			char sTmp[64];
			sprintf(sTmp, "sp%u/ab=%.1f/name=", Id, Ab);
			sLabel = string(sTmp) + Name;
			}

		const byte *Seq = ptrOutDB->GetSeq(SeqIndex);
		unsigned L = ptrOutDB->GetSeqLength(SeqIndex);
		if (fOut != 0)
			SeqToFasta(fOut, sLabel.c_str(), Seq, L);
		}
	CloseStdioFile(fOut);
	}

int main(int argc, char **argv)
	{
#if	defined(DEBUG) && defined(_MSC_VER)
	_CrtSetDbgFlag(_CRTDBG_CHECK_ALWAYS_DF);
#endif

	MyCmdLine(argc, argv);

	if (opt_version)
		{
		printf("chsim v" MY_VERSION ".%s\n", SVN_VERSION);
		return 0;
		}

	if (opt_svnmods)
		{
		printf("Revision %s\n", SVN_VERSION);
		printf("%s\n", SVN_MODS);
		return 0;
		}

	ChSim();

	ProgressExit();
	}
